'----------------------------------------------------------------------
'
'' Module: LCD_4P.bas
'
' Copyright 2006, 2008 Elba Corp.
' All rights reserved.
'
' Non-exclusive permission to use, copy, modify, and distribute this
' software for any purpose and without fee is hereby granted, provided:
'   1) that the above copyright notice is retained in the source code and,
'   2) that it is understood and agreed that no warranty of any kind,
'      express or implied is provided.
' Use of this code in any manner shall be prima facie evidence of
' acceptance of these terms.
'
' You may contact the author by email at dkinzer@zbasic.net.
'
'----------------------------------------------------------------------
'
' This module provides functions to interact with a 2 line by 16 character,
' parallel data LCD with a standard Hitachi 44780 command set.  Two different
' interface modes are supported: a direct-wire method and using an I2C I/O
' expander.  In either case, the LCD is connected using 4 data lines and
' three control signals: RS, R/W and E.  If write-only access to the LCD
' is desired, the R/W signal may be wired to ground.
'
' Information on the Hitachi 44780 command set can be found at
' http://www.myke.com/lcd.htm and elsewhere on the net.
'
'----------------------------------------------------------------------

' LCD command values
Private Const LCD_INIT as Byte = &H28			' 4 bits, 2 lines, 5x7 cell
Private Const LCD_CLEAR as Byte = &H01			' clear the display memory
Private Const LCD_HOME as Byte = &H02			' move cursor to the home position
Private Const LCD_ADDR_INC as Byte = &H06		' increment cursor after writing
Private Const LCD_MODE_SET as Byte = &H08		' display on, blinking cursor
Private Const LCD_SET_ADDR as Byte = &H80		' add to the desired address

' additional flags for the LCD_MODE_SET command
Private Const LCD_DISPLAY_ON As Byte = &H04		' turn on display
Private Const LCD_CURSOR_ON As Byte = &H02		' turn on cursor
Private Const LCD_CURSOR_BLINK As Byte = &H01	' turn on display

Private Const LCD_BUSY as Byte = &H80			' indicates the LCD is busy

' These bitmask constants represent the control signals.  For the
' direct-wire interface, they must match the port.bit definitions
' given below.  In any event, they must match the connections of
' the corresponding signals to the LCD.
Private Const LCD_DAT as Byte = &H08			' the Register Select line
Private Const LCD_CMD as Byte = &H00
Private Const LCD_E as Byte = &H04				' the Enable line
Private Const LCD_RD as Byte = &H02				' the R/W* line
Private Const LCD_WR as Byte = &H00

Private Const LCD_DATA_MASK as Byte = &Hf0		' LCD data occupies the high 4 bits

' the LCD requires some time for initialization
Private Const LCD_POWER_ON_DELAY as Single = 0.050

' commands used in the initialization sequence
Private Const LCD_4BIT_INIT_CMD1 as Byte = &H30
Private Const LCD_4BIT_INIT_CMD2 as Byte = &H20

'----------------------------------------------------------------------
' configuration controls
'----------------------------------------------------------------------

' Uncomment the following line to configure for the I2C interface.
'#define LCD_I2C

' Uncomment the following line to configure for a 3-wire interface
' using the 74LS164 shift register.
'#define LCD_3WIRE

' Uncomment the following line to implement status checking.
' Otherwise, post-operation delays are used.  The latter method
' allows operation with the R/W line wired to logic zero.  This
' cannot be used with the 3-wire interface.
'#define CHECK_BUSY

'----------------------------------------------------------------------
' configuration-specific data
'----------------------------------------------------------------------

#if defined(LCD_I2C)

'----------------------------------------------------------------------
' Definitions specific to the I2C interface.
'----------------------------------------------------------------------

' define the I2C channel to use
Private Const chan as Byte = 0

' define the address components of the PCF8574
Private Const i2cAddrBase as Byte = &H40			' this part is fixed
Private Const i2cAddrOfst as Byte = &H00			' this part is selectable by A2-A0
Private Const i2cAddr as Byte = i2cAddrBase + i2cAddrOfst

#elseif defined(LCD_3WIRE)

'----------------------------------------------------------------------
' Definitions specific to the 3-wire interface using a 74LS164.
'----------------------------------------------------------------------

' LCD interface pins
Private Const pinE as Byte = A.0				' the LCD E signal
Private Const pinData as Byte = A.1				' shift register input data
Private Const pinClock as Byte = A.2			' shift register clock

Private Const ePulseWidth as Single = 5.0e-5

#else

'----------------------------------------------------------------------
' Definitions specific to the direct-wire interface.
'----------------------------------------------------------------------

' LCD interface pins
Private Const pinRS as Byte = A.3				' the LCD RS signal (low for commands, high for data)
Private Const pinE as Byte = A.2				' the LCD E signal
Private Const pinRW as Byte = A.1				' the LCD R/W signal (low for writing, high for reading)

' length of the E pulse
Private Const ePulseWidth as Single = 5.0e-5

#endif

'----------------------------------------------------------------------
'
'' DisplayClear
'
' Clear the display.
'
Public Sub DisplayClear()
	Call lcdSend(LCD_CLEAR, LCD_CMD)
End Sub

'----------------------------------------------------------------------
'
'' DisplaySetPos
'
' Set the cursor position to the given row and column, both
' of which are zero based.
'
Public Sub DisplaySetPos(ByVal row as Byte, ByVal col as Byte)
	Call lcdSend(LCD_SET_ADDR + ((row * 64) And &H40) + (col And &H3f), LCD_CMD)
End Sub

'----------------------------------------------------------------------
'
'' DisplayChar
'
Public Sub DisplayChar(ByVal c as Byte)
	Call lcdSend(c, LCD_DAT)
End Sub

'----------------------------------------------------------------------
'
'' DisplayCharAt
'
Public Sub DisplayCharAt(ByVal c as Byte, ByVal row as Byte, ByVal col as Byte)
	Call LockTask()
	Call DisplaySetPos(row, col)
	Call lcdSend(c, LCD_DAT)
	Call UnlockTask()
End Sub

'----------------------------------------------------------------------
'
'' DisplayStr
'
' Display a string at the current position.
'
Public Sub DisplayStr(ByVal str as String)
	Dim cnt as Integer
	Dim idx as Integer

	cnt = Len(str)
	For idx = 1 to cnt
		Call DisplayChar(Asc(str, idx))
	Next idx
End Sub

'----------------------------------------------------------------------
'
'' DisplayStrAt
'
' Display a string at a specified position.
'
Public Sub DisplayStrAt(ByVal str as String, ByVal row as Byte, ByVal col as Byte)
	Call LockTask()
	Call DisplaySetPos(row, col)
	Call DisplayStr(str)
	Call UnlockTask()
End Sub

'----------------------------------------------------------------------
'
'' DisplayByteDec
'
Public Sub DisplayByteDec(ByVal bval as Byte)
	Call DisplayStr(CStr(bval))
End Sub

'----------------------------------------------------------------------
'
'' InitLCD
'
' Send initialization commands to the LCD.  This routine uses the
' method described at http://www.myke.com/lcd.htm to put the LCD
' in 4-bit mode.
'
Public Sub InitLCD()
	Dim i as Byte

#if defined(LCD_I2C)

	Dim data(1 to 3) as Byte
	Dim stat as Integer

	' it is assumed that the hardware I2C channel is being used
	Call OpenI2C(chan, 0, 0)

	' set the outputs of the I/O expander low
	data(1) = &H00
	stat = I2CCmd(chan, i2cAddr, 1, data, 0, 0)

	' delay for power-on initialization
	Call Delay(LCD_POWER_ON_DELAY)

	' put the display in 4-bit mode
	data(1) = LCD_4BIT_INIT_CMD1
	stat = I2CCmd(chan, i2cAddr, 1, data, 0, 0)
	data(1) = LCD_4BIT_INIT_CMD1 Or LCD_E
	data(2) = LCD_4BIT_INIT_CMD1
	For i = 1 to 3
		stat = I2CCmd(chan, i2cAddr, 2, data, 0, 0)
		Call PulseOut(0, 5.0e-3, 0)
	Next i
	data(1) = LCD_4BIT_INIT_CMD2
	data(2) = data(1) Or LCD_E
	data(3) = data(1)
	stat = I2CCmd(chan, i2cAddr, 3, data, 0, 0)

#elseif defined(LCD_3WIRE)
	Dim data as Byte

	' initialize the LCD interface pins
	Call PutPin(pinE, 0)
	Call PutPin(pinData, 0)
	Call PutPin(pinClock, 0)

	' delay for power-on initialization
	Call Delay(LCD_POWER_ON_DELAY)

	' put the display in 4-bit mode
	data = LCD_4BIT_INIT_CMD1
	Call ShiftOut(pinData, pinClock, 8, data)
	For i = 1 to 3
		Call PulseOut(pinE, ePulseWidth, 1)
		Call PulseOut(0, 5.0e-3, 0)
	Next i
	data = LCD_4BIT_INIT_CMD2
	Call ShiftOut(pinData, pinClock, 8, data)
	Call PulseOut(pinE, ePulseWidth, 1)

#else

	' initialize the LCD interface pins
	Register.PortA = Register.PortA And &H01
	Register.DDRA = Register.DDRA Or &Hfe

	' delay for power-on initialization
	Call Delay(LCD_POWER_ON_DELAY)

	' put the display in 4-bit mode
	Call SetBits(Register.PortA, LCD_DATA_MASK, LCD_4BIT_INIT_CMD1)
	For i = 1 to 3
		Call PulseOut(pinE, ePulseWidth, 1)
		Call PulseOut(0, 5.0e-3, 0)
	Next i
	Call SetBits(Register.PortA, LCD_DATA_MASK, LCD_4BIT_INIT_CMD2)
	Call PulseOut(pinE, ePulseWidth, 1)

#endif

	' 5mS delay
	Call PulseOut(0, 5.0e-3, 0)

	' send initialization data
	Call lcdSend(LCD_INIT, LCD_CMD)
	Call lcdSend(LCD_CLEAR, LCD_CMD)
	Call lcdSend(LCD_MODE_SET + LCD_DISPLAY_ON, LCD_CMD)
	Call lcdSend(LCD_ADDR_INC, LCD_CMD)
	Call lcdSend(LCD_SET_ADDR, LCD_CMD)
End Sub

'----------------------------------------------------------------------
'
'' lcdSend
'
' Send a command or data byte to the LCD.
'
Private Sub lcdSend(ByVal bval as Byte, ByVal mode as Byte)
	mode = mode And LCD_DAT

#if defined(CHECK_BUSY) And Not defined(LCD_3WIRE)
	' wait until the LCD is ready to accept data
	Call waitNotBusy()
#endif

#if defined(LCD_I2C)

	Dim data(1 to 6) as Byte
	Dim stat as Integer

	' prepare the I2C data stream
	data(1) = (bval And LCD_DATA_MASK) Or mode
	data(2) = data(1) Or LCD_E
	data(3) = data(1)
	data(4) = (Shl(bval, 4) And LCD_DATA_MASK) Or mode
	data(5) = data(4) Or LCD_E
	data(6) = data(4)
	stat = I2CCmd(chan, i2cAddr, 6, data, 0, 0)

#elseif defined(LCD_3WIRE)
	Dim data as Byte

	' send the high 4 bits
	data = (bval And LCD_DATA_MASK) Or mode
	Call ShiftOut(pinData, pinClock, 8, data)
	Call PulseOut(pinE, ePulseWidth, 1)

	' send the low 4 bits
	data = (Shl(bval, 4) And LCD_DATA_MASK) Or mode
	Call ShiftOut(pinData, pinClock, 8, data)
	Call PulseOut(pinE, ePulseWidth, 1)

#else

	' set the control lines
	Call SetBits(Register.PortA, &H0e, mode)

	' send the high 4 bits
	Call SetBits(Register.PortA, LCD_DATA_MASK, bval And LCD_DATA_MASK)
	Call PulseOut(pinE, ePulseWidth, 1)

	' send the low 4 bits
	Call SetBits(Register.PortA, LCD_DATA_MASK, Shl(bval, 4) And LCD_DATA_MASK)
	Call PulseOut(pinE, ePulseWidth, 1)

#endif

#if Not defined(CHECK_BUSY) Or defined(LCD_3WIRE)
	' implement post-command delay
	If (Not CBool(mode And LCD_DAT)) Then
		If (CBool(bval And (LCD_CLEAR Or LCD_HOME))) Then
			Call PulseOut(0, 5.0e-3, 0)	' 5mS delay
		Else
			Call PulseOut(0, 0.2e-3, 0)	' 200uS delay
		End if
	End If
#endif

End Sub

#if defined(CHECK_BUSY) And Not defined(LCD_3WIRE)

'----------------------------------------------------------------------
'
'' waitNotBusy
'
' Wait for the LCD to indicate that it is ready to accept a command or
' data value.
'
Private Sub waitNotBusy()
#ifdef LCD_I2C

	Dim data as Byte
	Dim stat as Integer

	data = LCD_RD Or LCD_CMD Or LCD_DATA_MASK
	stat = I2CCmd(chan, i2cAddr, 1, data, 0, 0)
	Do
		Dim res as Byte

		' assert the Enable signal, read the status
		data = data Or LCD_E
		stat = I2CCmd(chan, i2cAddr, 1, data, 1, res)

		' deassert the Enable signal
		data = data And Not LCD_E
		stat = I2CCmd(chan, i2cAddr, 1, data, 0, 0)
	Loop While CBool(res And LCD_BUSY)
	data = 0
	stat = I2CCmd(chan, i2cAddr, 1, data, 0, 0)

#else

	' set the data pins to be inputs
	Register.DDRA = Register.DDRA And &H0f
	Register.PortA = Register.PortA And &H0f
	Call SetBits(Register.PortA, &H0e, LCD_RD Or LCD_CMD)
	Do
		Dim stat as Byte

		' read the status
		Call PutPin(pinE, 1)
		stat = Register.PinA
		Call PutPin(pinE, 0)
		Call PulseOut(pinE, 1, 1)
	Loop While CBool(stat And LCD_BUSY)
	Call SetBits(Register.PortA, &H0e, 0)
	
	' set the pins to be outputs again
	Register.PortA = Register.PortA And &H0F
	Register.DDRA = Register.DDRA Or &Hf0

#endif
End Sub

#endif

'----------------------------------------------------------------------

' Uncomment this line to include a simple test driver.
#define TEST_DRIVER

#if defined(TEST_DRIVER)

Sub Main()
	Call InitLCD()
	Call DisplayStrAt("Hello,", 0, 5)
	Call DisplayStrAt("world!", 1, 5)
End Sub

#endif
